%{
#include "at.h"
#include "config.h"
#include "y.tab.h"
#include <ctype.h>

#ifdef FLEX_SCANNER
/*
 * Redefine the input macro to get input from the at string.  Gets
 * input and stuffs it into "buf", returning the number of characters
 * read in "result".
 */
#define YY_INPUT(buf, result, max_size) \
    do \
	{ \
	for (result = 0;  result < max_size;  result++, atstr++) \
	    if ((buf[result] = *atstr) == '\0') \
		break; \
	} \
	while (0)
#else /* FLEX_SCANNER */
/*
 * Redefine the getc () macro and the feof macro.  This is actually the only
 * way to change the input stream, because the generated code runs feof
 * on yyin, rather than using a macro like the 'input' macro.
 */
#undef getc
#define getc(x)	((*atstr == '\0') ? EOF : *atstr++)
#undef feof
#define feof(x)	(*atstr == '\0')
#endif /* FLEX_SCANNER */

struct keylist
    {
    char *	kl_word;	/* String to match */
    int		kl_value;	/* Value to be returned as scanner value */
    int		kl_lval;	/* Value to be returned in yylval */
    };

struct keylist	mainkeys[] =
		    {
		    "now",	K_NOW,		0,
		    "next",	K_NEXT,		0,
		    "am",	K_AM_PM,	0,
		    "pm",	K_AM_PM,	1,
		    "zulu",	K_AM_PM,	2,
		    "gmt",	K_AM_PM,	2,
		    "greenwich", K_AM_PM,	2,
		    "utc",	K_AM_PM,	2,
		    "january",	K_MONTH,	1,
		    "february",	K_MONTH,	2,
		    "march",	K_MONTH,	3,
		    "april",	K_MONTH,	4,
		    "may",	K_MONTH,	5,
		    "june",	K_MONTH,	6,
		    "july",	K_MONTH,	7,
		    "august",	K_MONTH,	8,
		    "september",K_MONTH,	9,
		    "october",	K_MONTH,	10,
		    "november",	K_MONTH,	11,
		    "december",	K_MONTH,	12,
		    "jan",	K_MONTH,	1,
		    "feb",	K_MONTH,	2,
		    "mar",	K_MONTH,	3,
		    "apr",	K_MONTH,	4,
		    "may",	K_MONTH,	5,
		    "jun",	K_MONTH,	6,
		    "jul",	K_MONTH,	7,
		    "aug",	K_MONTH,	8,
		    "sep",	K_MONTH,	9,
		    "oct",	K_MONTH,	10,
		    "nov",	K_MONTH,	11,
		    "dec",	K_MONTH,	12,
		    "today",	K_RELDAY,	0,
		    "tomorrow",	K_RELDAY,	1,
		    "monday",	K_WEEKDAY,	1,
		    "tuesday",	K_WEEKDAY,	2,
		    "wednesday",K_WEEKDAY,	3,
		    "thursday",	K_WEEKDAY,	4,
		    "friday",	K_WEEKDAY,	5,
		    "saturday",	K_WEEKDAY,	6,
		    "sunday",	K_WEEKDAY,	0,
		    "mon",	K_WEEKDAY,	1,
		    "tue",	K_WEEKDAY,	2,
		    "wed",	K_WEEKDAY,	3,
		    "thu",	K_WEEKDAY,	4,
		    "fri",	K_WEEKDAY,	5,
		    "sat",	K_WEEKDAY,	6,
		    "sun",	K_WEEKDAY,	0,
		    "noon",	K_NOON,		0,
		    "midnight",	K_MIDNIGHT,	0,
		    "midnite",	K_MIDNIGHT,	0,
		    "minute",	K_INTERVAL,	I_MINUTE,
		    "minutes",	K_INTERVAL,	I_MINUTE,
		    "min",	K_INTERVAL,	I_MINUTE,
		    "mins",	K_INTERVAL,	I_MINUTE,
		    "hour",	K_INTERVAL,	I_HOUR,
		    "hours",	K_INTERVAL,	I_HOUR,
		    "hr",	K_INTERVAL,	I_HOUR,
		    "hrs",	K_INTERVAL,	I_HOUR,
		    "day",	K_INTERVAL,	I_DAY,
		    "days",	K_INTERVAL,	I_DAY,
		    "dy",	K_INTERVAL,	I_DAY,
		    "dys",	K_INTERVAL,	I_DAY,
		    "week",	K_INTERVAL,	I_WEEK,
		    "weeks",	K_INTERVAL,	I_WEEK,
		    "wk",	K_INTERVAL,	I_WEEK,
		    "wks",	K_INTERVAL,	I_WEEK,
		    "month",	K_INTERVAL,	I_MONTH,
		    "months",	K_INTERVAL,	I_MONTH,
		    "mo",	K_INTERVAL,	I_MONTH,
		    "mos",	K_INTERVAL,	I_MONTH,
		    "mon",	K_INTERVAL,	I_MONTH,
		    "mons",	K_INTERVAL,	I_MONTH,
		    "mth",	K_INTERVAL,	I_MONTH,
		    "mths",	K_INTERVAL,	I_MONTH,
		    "year",	K_INTERVAL,	I_YEAR,
		    "years",	K_INTERVAL,	I_YEAR,
		    "yr",	K_INTERVAL,	I_YEAR,
		    "yrs",	K_INTERVAL,	I_YEAR,
		    0,		0,		0
		    };

extern char *	atstr;		/* Characters to be parsed */
extern char *	where;
extern int	yylval;		/* Lex return value */
%}

%%
[0-9]*			{
			yylval = atoi (yytext);
			return (strlen (yytext) > 2) ? K_BIG_NUMBER : K_NUMBER;
			}

[a-zA-Z_][0-9a-zA-Z_]*	{
			reg2 char *		cp;
			reg1 struct keylist *	keyp;

			for (cp = yytext;  *cp != '\0';  cp++)
			    if (isupper (*cp))
				*cp = tolower (*cp);
			for (keyp = mainkeys;  keyp->kl_word;  keyp++)
			    if(strcmp(keyp->kl_word, yytext) == 0)
				{
				yylval = keyp->kl_lval;

				return keyp->kl_value;
				}
			return K_SYNTAX;
			}
"@"[^ \t]+		{
			if (where)
				free(where);
			where = strdup(yytext+1);
			return K_PLACE;
			}
[:+,]			return yylval = yytext[0];

[ \t]			;

.			return K_SYNTAX;

%%
int yywrap ()
    {
    return 1;
    }
