#ifndef lint
#ifdef GEN_ID
static char cron_h_Id[] =
    "$Header$";
#endif
#endif

/*
 * $Log$
 * Revision 2.20  2000/04/13 03:19:35  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.19  2000/03/23 00:59:06  geoff
 * Change the Y2K threshold to 100, since the 1900's are over.
 *
 * Revision 2.18  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.17  1998/07/30 20:42:23  geoff
 * Many changes, accumulated over far too much time: add longname (Y2K)
 * support and support for resource-exhaustion limits.
 *
 * Revision 2.16  1995/11/27  01:59:28  geoff
 * Move the definitions of HOUR and MINUTE.  Add MAXSLEEP and MINUPDLAST.
 * Change all the debugging macros to preserve errno during output.  Add
 * the EF_ONLYCATCHUP flag.
 *
 * Revision 2.15  1995/05/11  05:05:09  geoff
 * Minor RCS header changes.
 *
 * Revision 2.14  87/01/10  19:14:22  geoff
 * Add symbols for the return values from sendcommand.
 * 
 * Revision 2.13  86/04/17  21:55:04  geoff
 * Add a compile-able RCS ID.
 * 
 * Revision 2.12  86/04/02  00:01:08  geoff
 * Move all the configuration stuff to config.h.
 * 
 * Revision 2.11  86/03/31  22:57:20  geoff
 * Add structures and constants to support limiting the size of the process
 * queues, and to support notifying the user of limit problems.  Also remove
 * the signal-based dumping stuff since tellcron does it more cleanly.
 * 
 * Revision 2.10  86/03/30  15:38:47  geoff
 * Add DBPRINTn macros
 * 
 * Revision 2.9  86/03/30  14:32:56  geoff
 * Add HUGETIME.
 * 
 * Revision 2.8  86/03/28  14:40:49  geoff
 * Add code to handle debugging options better.  Add more fields and
 * definitions to support user limits.  Add support for the dump signal.  Add
 * ce_rtime, and the "has run" flag.  Fix some lint complaints.
 * 
 * Revision 2.7  86/03/25  19:55:29  geoff
 * Add several symbols, mostly useful to implementing 'at', but a few for
 * bugs/new features.
 * 
 * Revision 2.6  86/01/04  16:15:48  geoff
 * Begin preparations for an at command.  Add command queues to the data
 * structures.  Add vfork, DAEMONUID, and DAEMONGID.
 * 
 * Revision 2.5  85/11/25  16:51:09  geoff
 * Change FIFOTIME to be ERRORTIME (it's more generic).
 * 
 * Revision 2.4  85/11/24  16:11:54  geoff
 * Rearrange file naming to be relative to current directories;  the routines
 * that open files are responsible for cd'ing to CRONTABDIR if necessary and
 * leaving cron in LIBDIR afterwards.  Add CRONLOCK and LOCKLINK, CHARMASK,
 * READFIFO, FIFOTIME.  Delete EF_NEWFILES and EF_USERTAB.
 * 
 * Revision 2.3  85/11/05  00:35:34  geoff
 * Move the childproc structure here from runcmd.c, and change all flag names
 * to reflect the structure they apply to.
 * 
 * Revision 2.2  85/11/02  19:06:52  geoff
 * Add variables needed by the runcmd.c changes to mail errors to the user.
 * 
 * Revision 2.1  85/10/25  02:04:32  geoff
 * Add support for new format crontabs and symbolic names
 * 
 * Revision 1.2  85/10/13  21:44:50  geoff
 * Add FIFO type selection parameters
 * 
 * Revision 1.1  85/10/12  01:33:31  geoff
 * Break it up into separate modules
 * 
 */

#include "config.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

/*
 * Selection of signals for communications with the cron daemon
 */
#ifdef SIGUSR1
#define READFIFO	SIGUSR1		/* Signal that says to read the fifo */
#else
#define READFIFO	SIGSYS		/* Signal that says to read the fifo */
#endif

/*
 * Sizes of certain environment variables
 */
#define EL_HOME		(sizeof ENVHOME - 1)
#define EL_LOGNAME	(sizeof ENVLOGNAME - 1)
#define EL_TZ		(sizeof ENVTZ - 1)

/*
 * Some UNIX systems don't have symbolic values for the permission
 * bits in their stat.h files.  So we'll have to do it ourselves.
 */
#ifndef S_IEXEC
#define	S_IEXEC		00100
#define	S_IWRITE	00200
#define	S_IREAD		00400
#define	S_ISVTX		01000
#define	S_ISGID		02000
#define	S_ISUID		04000
#endif

/*
 * Return values from "sendcommand".
 */
#define SEND_OK		0		/* Command was sent successfully */
#define SEND_QUEUED	1		/* No daemon, command was queued */
#define SEND_ERROR	2		/* Errors sending the command */

/*
 * Required file modes for crontab files.  If a file does not satisfy
 * these modes, it will not be accepted as a crontab.
 *
 * File mode must have SUID, SGID, and IREAD, but not SVTX, any EXEC, or
 * world write permissions.
 */
#define MUSTHAVE	(S_ISUID | S_ISGID | S_IREAD)
#define CANTHAVE	(S_ISVTX | S_IEXEC | (S_IEXEC >> 3) \
			  | ((S_IEXEC | S_IWRITE) >> 6))

/*
 * Default modes for crontab files received on standard input
 */
#define CRONMODE	S_IREAD		/* Very paranoid */

/*
 * Default modes for 'at' files received on standard input
 */
#define ATMODE		S_IREAD		/* Very paranoid */

#ifndef lint
#ifndef LOGBITSPERMASK
#if BITSPERMASK == 4
#define LOGBITSPERMASK	2
#endif
#if BITSPERMASK == 8
#define LOGBITSPERMASK	3
#endif
#if BITSPERMASK == 16
#define LOGBITSPERMASK	4
#endif
#if BITSPERMASK == 32
#define LOGBITSPERMASK	5
#endif
#if BITSPERMASK == 64
#define LOGBITSPERMASK	6
#endif
#endif
#endif

#define HOUR	(60 * 60)		/* Number of seconds in an hour */
#define MINUTE	60			/* Number of seconds in a minute */
#define CATCHUP	(3 * MINUTE)		/* Catch up if time changes by more */
					/* ..than this much */
#define ERRORTIME (10 * MINUTE)		/* Interval between error complaints */
#define MAXSLEEP (5 * MINUTE)		/* Maximum time to ever sleep */
#define MINUPDLAST ((time_t) (1 * HOUR)) /* Update last-run record at least */
					/* ..this often */

/*
 * If a request is more than this many years in the future, it is not put
 * into the tables.  The assumption is that cron will be restarted
 * sometime during the next YEARMAX years.
 */
#ifdef lint
#define YEARMAX	(BITSPERMASK)
#else
#if BITSPERMASK < 6
#define YEARMAX	((6 + BITSPERMASK - 1 / BITSPERMASK) * BITSPERMASK)
#else
#define YEARMAX	(BITSPERMASK)
#endif
#endif

/*
 * Protopytes for and offsets into the "when" part of at-job names
 */
#ifdef LONGNAMES
#define WHEN_PROTOTYPE "a19991231235959XXXXXXXXXXXX"
#define WHEN_TRAILER_PROTOTYPE "XXXXXXXXXXXX"
#define WHEN_FLAG	0	/* Flag character */
#define WHEN_FLAG_VALUE	'a'
#define WHEN_YEAR	1	/* 4-digit year */
#define WHEN_MONTH	5	/* 2-digit month */
#define WHEN_DAY	7	/* 2-digit day of month */
#define WHEN_HOUR	9	/* 2-digit hour */
#define WHEN_MINUTE	11	/* 2-digit minute */
#define WHEN_ID		13	/* 2-digit current second for ID purposes */
#define WHEN_X_TRAILER	15	/* Trailing X's for uniquifier and PID */
#define WHEN_OLD_PROTOTYPE "991231235959XXXXXXX"
#define WHEN_OLD_YEAR	0	/* 2-digit year */
#define WHEN_OLD_MONTH	2	/* 2-digit month */
#define WHEN_OLD_DAY	4	/* 2-digit day of month */
#define WHEN_OLD_HOUR	6	/* 2-digit hour */
#define WHEN_OLD_MINUTE	8	/* 2-digit minute */
#define WHEN_OLD_ID	10	/* 2-digit current second for ID purposes */
#define WHEN_OLD_X_TRAILER 12	/* Trailing X's for uniquifier and PID */
#else /* LONGNAMES */
#define WHEN_PROTOTYPE "991231235959XXXXXXX"
#define WHEN_TRAILER_PROTOTYPE "XXXXXXX"
#define WHEN_YEAR	0	/* 2-digit year */
#define WHEN_MONTH	2	/* 2-digit month */
#define WHEN_DAY	4	/* 2-digit day of month */
#define WHEN_HOUR	6	/* 2-digit hour */
#define WHEN_MINUTE	8	/* 2-digit minute */
#define WHEN_ID		10	/* 2-digit current second for ID purposes */
#define WHEN_X_TRAILER	12	/* Trailing X's for uniquifier and PID */
#endif /* LONGNAMES */

#ifdef LOC_CRON
#ifdef LOC_ATLOC
#define ATLOC_NOYEAR 1971 /* jobs with this year are edge triggered */
#define ATLOC_FLAG_VALUE 'b'
#define ATLOC_PROTOTYPE "bXXXXXXXXXXXX"
#define ATLOC_TRAILER_PROTOTYPE "XXXXXXXXXXXX"
#define ATLOC_X_TRAILER	1	/* Trailing X's for uniquifier and PID */
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */

/*
 * Year-2000 information for dealing with old-format 'at' jobs.  The
 * threshold is only 2 digits, since it only arises in situations with
 * 2-digit years.
 */
#define Y2K_THRESHOLD	100	/* 1900's are over: 2-digit years become 2K */
#define Y2K_1900	1900	/* Just in case we need to change in 2100 */
#define Y2K_2000	2000	/* Just in case we need to change in 2100 */

/*
 * Helpful macros for dealing with bit fields
 */
#ifdef LOGBITSPERMASK
#define BITSIZE(x)	(((x) + BITSPERMASK - 1) >> LOGBITSPERMASK)
#define MASKBYTE(x)	((x) >> LOGBITSPERMASK)
#define MASKBIT(x)	(1 << ((x) & (BITSPERMASK - 1)))
#else
#define BITSIZE(x)	(((x) + BITSPERMASK - 1) / BITSPERMASK)
#define MASKBYTE(x)	((x) / BITSPERMASK)
#define MASKBIT(x)	(1 << ((x) % BITSPERMASK))
#endif

/*
 * Debugging printf macros.
 */
#ifdef DEBUG
extern short	debug;
#define DBPRINT0(lvl, fmt) \
			if (debug >= lvl) \
			    { \
			    int saveerr = errno; \
			    (void) fseek (stderr, 0L, 2); \
			    errno = saveerr; \
			    (void) fprintf (stderr, fmt); \
			    } \
			else
#define DBPRINT1(lvl, fmt, arg1) \
	    		if (debug >= lvl) \
			    { \
			    int dbsaveerr = errno; \
			    (void) fseek (stderr, 0L, 2); \
			    errno = dbsaveerr; \
			    (void) fprintf (stderr, fmt, arg1); \
			    } \
			else
#define DBPRINT2(lvl, fmt, arg1, arg2) \
	    		if (debug >= lvl) \
			    { \
			    int dbsaveerr = errno; \
			    (void) fseek (stderr, 0L, 2); \
			    errno = dbsaveerr; \
			    (void) fprintf (stderr, fmt, arg1, arg2); \
			    } \
			else
#define DBPRINT3(lvl, fmt, arg1, arg2, arg3) \
	    		if (debug >= lvl) \
			    { \
			    int dbsaveerr = errno; \
			    (void) fseek (stderr, 0L, 2); \
			    errno = dbsaveerr; \
			    (void) fprintf (stderr, fmt, arg1, arg2, arg3); \
			    } \
			else
#define DBPRINT4(lvl, fmt, arg1, arg2, arg3, arg4) \
	    		if (debug >= lvl) \
			    { \
			    int dbsaveerr = errno; \
			    (void) fseek (stderr, 0L, 2); \
			    errno = dbsaveerr; \
			    (void) fprintf (stderr, fmt, \
			      arg1, arg2, arg3, arg4); \
			    } \
			else
#else
#define DBPRINT0(lvl, fmt)
#define DBPRINT1(lvl, fmt, arg1)
#define DBPRINT2(lvl, fmt, arg1, arg2)
#define DBPRINT3(lvl, fmt, arg1, arg2, arg3)
#define DBPRINT4(lvl, fmt, arg1, arg2, arg3, arg4)
#endif

/*
 * The following structures (cronuser, cronfile, cronent, and cmdqueue)
 * contain mutual pointers to each other, in a circular fashion.  In an
 * attempt to make this compile everywhere, I am trying a new technique:
 * nested structure definitions.  If you have trouble with it, I'd sure
 * like to find out.
 */

/*
 * Layout of the internal crontab entry
 */
struct cronent
    {
    struct cronent *
		ce_next;		/* Next entry on crontab list */
    struct cronent *
		ce_prev;		/* Previous entry on crontab list */
    int		ce_uid;			/* User ID to run under */
    int		ce_gid;			/* Group ID to run under */
    short	ce_flags;		/* Flag bits, see below */
    MASKTYPE	ce_min[BITSIZE (60)];	/* Minutes */
    MASKTYPE	ce_hour[BITSIZE (24)];	/* Hours */
    MASKTYPE	ce_moday[BITSIZE (31)];	/* Day of month */
    MASKTYPE	ce_month[BITSIZE (12)];	/* Month of year */
    MASKTYPE	ce_wkday[BITSIZE (7)];	/* Day of week */
    MASKTYPE	ce_year[BITSIZE (YEARMAX)]; /* Year */
    time_t	ce_rtime;		/* Next time entry is due to run */
#ifdef LOC_CRON
    char *      ce_location;          /* Location */
#ifdef LOC_EVERY
    char *      ce_every_spec;		/* spec for how frequently to run job */
    time_t      ce_every_next;          /* next possible time to run */
#endif /* LOC_EVERY */
#endif
    char *	ce_command;		/* Command to run */
    char *	ce_input;		/* Input to give to it (if any) */
    /*
     * Structure to describe a file containing a cron table
     */
    struct cronfile
	{
	struct cronfile *
		    cf_next;		/* Next entry on cronfile list */
	struct cronfile *
		    cf_prev;		/* Previous entry on cronfile list */
	char *	    cf_name;		/* Pathname of the file */
	time_t	    cf_rtime;		/* When we last read this file */
	short	    cf_flags;		/* See below */
	/*
	 * Structure to describe a user who has cron, at, or batch requests
	 */
	struct cronuser
	    {
	    struct cronuser *
			cu_next;	/* Next entry on user list */
	    struct cronuser *
			cu_prev;	/* Previous entry on user list */
	    int		cu_uid;		/* User's ID */
	    int		cu_gid;		/* User's group ID */
	    char *	cu_home;	/* User's home directory */
	    char *	cu_name;	/* User's name (for LOGNAME) */
	    unsigned long cu_entsizelim; /* Limit on entry size */
	    unsigned long cu_mailsizelim; /* Limit on mail size */
	    short	cu_flags;	/* Flags:  see below */
	    short	cu_entlim;	/* Limit on cron entries */
	    short	cu_procs;	/* Subprocesses this user has */
	    short	cu_proclim;	/* Limit on user's subprocesses */
	    short	cu_qcount;	/* Number of processes queued */
	    short	cu_qlim;	/* Limit on waiting process queue */
	    short	cu_nice;	/* Nice level to run user's stuff at */
	    /*
	     * Queue of commands awaiting execution.
	     */
	    struct cmdqueue
		{
		struct cmdqueue *
			    cq_next;	/* Next entry on run queue */
		struct cmdqueue *
			    cq_prev;	/* Previous entry on run queue */
		struct cronent *
			    cq_ent;	/* Entry to be run */
		short	    cq_flags;	/* Flags;  see below */
	    /* Back to cronuser */
		}
			cu_queue;	/* Command queue */
	/* Back to cronfile */
	    } *
			cf_user;	/* User who owns this file */
    /* Back to cronent */
	} *
		ce_file;		/* Cron file the entry came from */
    };

/*
 * Entry flags
 */
#define EF_WAIT		0x0001		/* Run during startup & wait for it */
#define EF_SPAWN	0x0002		/* Run during startup, don't wait */
#define EF_SEQSPAWN	0x0004		/* Run in seq during bckgrnd startup */
#define EF_RUNALL	0x0008		/* Run all catchups of this cmd */
#define EF_ONLYCATCHUP	0x0010		/* Run only during catchup mode */
#define EF_MARKDEL	0x0020		/* Entry is marked for final delete */
#define EF_DELETED	0x0040		/* Entry is logically deleted */
#define EF_ATJOB	0x0080		/* Entry is an at job, not cron */
#define EF_HASRUN	0x0100		/* Job has run during catchup */
#ifdef LOC_CRON
#define EF_IS_AT_LOCATION 0x0200	/* job is at suitable location to run */
#ifdef LOC_MOVE
#define EF_WAS_AT_LOCATION 0x0400	/*  " " ...before last move */
#define EF_ONLY_ON_MOVE 0x0800		/* should this job run only on moves? */
#define EF_ONLY_ON_MOVE_FIRED 0x1000     /* are at jobs already running? */
#endif /* LOC_MOVE */
#ifdef LOC_EVERY
#define EF_RUN_EVERY	0x0400		/* job should be run only every n time units */
#endif /* LOC_EVERY */
#endif /* LOC_CRON */

/*
 * File table flags
 */
#define FF_MARKDEL	0x01		/* Marked for delete */
#define FF_DELETED	0x02		/* Logically deleted */
#define FF_ATJOB	0x04		/* Entry is an at job, not cron file */

/*
 * User table flags
 */
#define UF_MARKDEL	0x0001		/* User is marked for delete */
#define UF_NEVERDEL	0x0002		/* Never delete user */
#define UF_LOSTJOBS	0x0004		/* Cron or at jobs were lost */

/*
 * Command queue flags
 */
#define QF_RANLATE	0x0001		/* Job was run late due to queuing */

/*
 * Structure to describe a process we have spawned
 */
struct childproc
    {
    struct childproc *
		cp_next;		/* Next entry on process list */
    struct childproc *
		cp_prev;		/* Previous entry on process list */
    struct cronent *
		cp_ent;			/* Entry that spawned the process */
    int		cp_pid;			/* Process ID of the process */
    short	cp_oprefix;		/* Size of stdout prefix */
    short	cp_eprefix;		/* Size of stdout prefix */
    char	cp_fbase[TEMPSIZE];	/* Base name of temp files for proc */
    };
