#!/usr/bin/perl -w

#
# power_monitor.pl
# $Id: power_monitor.pl 2908 2000-06-29 03:41:26Z johnh $
#
# Copyright (C) 1999 by John Heidemann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#



sub usage {
    print STDERR <<END;
usage: $0 [-D]

map linux apmd-3.0beta9 events into lcron events
monitor and filter location, sending signals to lcron if it changes
See apmd(8) or the source code of this script for details.

To install this, copy it to /etc/apmd/apmd_proxy (or just
/etc/apmd_proxy on RedHat 6.1).

END
    exit 1;
}

use strict;

use Getopt::Long;
&usage if ($#ARGV >= 0 && $ARGV[0] eq '-?');
my(%opts);
&GetOptions(\%opts, qw(d));
&usage if ($#ARGV < 0);

my($BINDIR) = '/usr/local/lib/xcron/bin';
my($LIBDIR) = '/usr/local/lib/xcron/lib';

my(%pass_commands) = qw(start pass
			stop pass
			suspend pass
			standby drop
			resume pass
			change check);

my($command) = $ARGV[0];
if (defined($pass_commands{$command}) && $pass_commands{$command} ne 'drop') {
    if ($command eq 'change' && $ARGV[1] eq 'power') {
	open(APM, "</proc/apm") || die "$0: cannot open /proc/apm\n";
	my($apm) = <APM>;
	my(@apm) = split(/ /, $apm);
	$command = ($apm[3] =~ /1/) ? "ac" : "battery";
	close APM;
    };
    my($fn) = "$LIBDIR/events/power";
    open(POWER, ">$fn") || die "$0: cannot open $fn\n";
    print POWER "power:$command\n";
    close POWER;
    chmod 0644, $fn || die "$0: cannot chmod $fn\n";
    &tellxcron;
};
exit 0;

sub tellxcron {
    system("$BINDIR/tellxcron", "-L");
}

#
# documentation of commands
# (this text is straight from apmd(8)):
# 
#        Most uses of this daemon will use  the  proxy  command  to
#        support  power  conservation  activities.  This command is
#        searched for by default  in  /etc/apmd/apmd_proxy  and  is
#        invoked with one or two arguments:
# 
#        start  Invoked when the daemon starts.  Normally sets sys-
#               tem-wide power  policy,  such  as  IDE  hard  drive
#               standby times, to account for whether battery power
#               is in use.
# 
#        stop   Invoked when the daemon stops.  Normally undoes any
#               policy settings done when the daemon started.
# 
#        suspend [ system | user ]
#               Invoked  when  the  APM  driver reports that system
#               suspension has been initiated.  The second  parame-
#               ter  indicates  whether  the  BIOS or a user action
#               (such as closing a laptop) initiated suspension.
# 
#               The  BIOS  "suspend"  mode  aggressively  conserves
#               power,  and normally involves shutting off power to
#               all devices except the CPU core and  memory,  which
#               is kept in a very low power mode.  Most laptops can
#               stay suspended, using battery power alone, for sev-
#               eral  days.  ("Hibernation" is a kind of super-sus-
#               pend, where all that state is written to  disk  and
#               the machine uses even less power bcause it can turn
#               off that CPU core, using no battery power  at  all.
#
#               At  this  writing,  Linux does not support hiberna-
#               tion.)  PCMCIA devices should be manually suspended
#               using cardctl(8), and some modular drivers may need
#               to be unloaded.
# 
#        standby [ system | user ]
#               Invoked when the APM  driver  reports  that  system
#               standby  has  been initiated.  The second parameter
#               indicates whether the BIOS or a user  action  (such
#               as invoking apm -s) caused this.
# 
#               The  BIOS  "standby" mode slightly conserves power,
#               and leaves the machine able to respond almost imme-
#               diately  to user activity.  Most laptops can't stay
#               in standby mode for even a day, if they  must  rely
#               on  battery  power.   Normally, nothing needs to be
#               done beyond what the BIOS itself will do.
# 
#        resume [ suspend | standby | critical ]
#               Invoked when the APM driver reports that system has
#               resumed  normal  operation.   The  second parameter
#               indicates what sort  of  mode  it  was  in  before,
#               either  the  "suspend"  mode (possibly a "critical"
#               suspend) or else "standby" mode.
# 
#               The system clock must be updated to match the hard-
#               ware clock; this will normally have been handled by
#               the kernel's APM driver.  PCMCIA devices  may  need
#               to  be  manually  resumed  from standby using card-
#               ctl(8), and some modular drivers  may  need  to  be
#               reloaded  or  otherwise reinitialized.  In the case
#               of a critical suspend, system state  may  not  have
#               been completely saved due to an emergency shutdown;
#               applications and and drivers may be in  a  confused
#               state.
# 
#        change power
#               This  presents  a  subset  of the APM driver "power
#               change" events, specifically those where  AC  power
#               was  added  or removed.  This will often modify the
#               system wide power policy; for example, so that  IDE
#               hard  drives  aggressively  enter standby mode when
#               only battery power is available.
# 
#        change battery
#               The APM driver has reported that  the  BIOS  thinks
#               the  strength  of one (or more) batteries is "low";
#               at least ten minutes of should power remain.
# 
#        change capability
#               Some change in the power management capabilities of
#               the  system  was reported.  It may have been caused
#               by operation of  some  setup  utility,  or  by  the
#               arrival or removal of some devices.
# 
# #
