#ifndef lint
static char Rcs_Id[] =
	"$Header$";
#endif

/*
 * $Log$
 * Revision 2.27  2000/02/08 08:56:42  geoff
 * Integrate various changes from Geoff's branch into the now-master CVS
 * repository.  As part of this integration, I cleaned up some, but not
 * all, of the inconsistent coding style introduced by the LOC_CRON
 * changes.  Important and substantive changes include:
 *
 *     Configure	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     Makefile.SH	Tiny cleanup to an expr command.
 *     config.SH	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     errmsg.c	Add errlist_in_stdio support.
 *     mainloop.c	Use setjmp/longjmp to ensure that we are blown off the
 * 		select when we catch an alarm, even if we are running
 * 		on a system that has reliable signals.  Also add
 * 		reliable-signal support that properly resets the
 * 		signal mask in the longjmp case.
 *     parse.c	Fix a bug that could cause crontabs to be accepted
 * 		with the wrong UID.  Be a bit more careful about
 * 		parsing years.
 *     runcmd.c	Add setjmp/longjmp and reliable-signal support.  (See
 * 		mainloop.c).
 *     xcron.c	Use setitimer to make absolutely sure that all alarms
 * 		are disabled at startup.
 *
 * Revision 2.26  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.25  1999/04/22 01:42:29  johnh
 * geoff's quick bugfix for hanging during long-running jobs
 *
 * Revision 2.24  1998/07/30 20:42:31  geoff
 * Run with a restrictive umask to prevent people reading other people's
 * /tmp files.
 *
 * Revision 2.23  1995/11/27  01:59:55  geoff
 * Instead of ignoring HUP, INT, and QUIT, catch them and exit cleanly.
 * Preserve errno when putting out error messages.  Make the "some other
 * cron is running" message a bit more clear, so the user has a clue
 * about what to do.  Work around problems with Linux stdio that keep us
 * from opening /dev/null for input on stdin even after closing it (is
 * this really a gcc bug?).
 *
 * Revision 2.22  1995/05/11  05:06:04  geoff
 * Support variable signal types.  Use setjmp/longjmp to handle reliable
 * signals properly.  Make RCS revision logging a bit more robust.
 *
 * Revision 2.21  88/10/31  23:54:31  geoff
 * Correct the commentary describing the -s switch.
 * 
 * Revision 2.20  87/04/01  23:51:35  geoff
 * Put the revision level into the startup message in a cleaner fashion.
 * 
 * Revision 2.19  87/03/31  22:34:43  geoff
 * Use ndir.h instead of dir.h, and include pstring.h to make sure its
 * RCS ID winds up in the object file.
 * 
 * Revision 2.18  86/12/01  14:47:07  geoff
 * Fix some lint complaints.
 * 
 * Revision 2.17  86/11/29  20:30:09  geoff
 * Make another stab at getting the revision in the startup message
 * 
 * Revision 2.16  86/11/29  20:24:46  geoff
 * Add revision information to the startup message
 * 
 * Revision 2.15  86/11/29  19:55:11  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.  Also add support for systems
 * without setpgrp() and for Berkeley-style setpgrp() calls with arguments.
 * 
 * Revision 2.14  86/11/16  01:00:11  geoff
 * Fix some of the leadin commentary.  Add (untested) support for the BSD
 * TIOCNOTTY ioctl.  Add #ifdef's for relative-nice and absolute-nice systems.
 * Add some more error checking to the open of the NULLINPUT input file.
 * 
 * Revision 2.13  86/04/17  22:01:16  geoff
 * Add support for compile-able RCS ID's.
 * 
 * Revision 2.12  86/04/13  13:33:59  geoff
 * Use the directory library to open and read directories, instead of
 * spawning ls.
 * 
 * Revision 2.11  86/04/01  19:04:28  geoff
 * Get reftime from lastruntime and pass it to filelist.  Get current time
 * from catchup and pass it to cron.
 * 
 * Revision 2.10  86/03/31  23:00:31  geoff
 * Remove the signal-based dumping stuff.  Also move the cron-started
 * message so it comes out before the readlimits message.
 * 
 * Revision 2.9  86/03/30  14:35:26  geoff
 * Add a choice of intervals to the -s option.  Add catching of signals so
 * temp files can be unlinked.  Make changes to support the sorted entry
 * list.
 * 
 * Revision 2.8  86/03/28  14:44:11  geoff
 * Add support for the user limits file, for dumps on demand, and for
 * sorting the entry list by next run time.
 * 
 * Revision 2.7  86/03/25  19:58:16  geoff
 * Add support for at jobs, and clean a couple of minor deficiencies.
 * 
 * Revision 2.6  86/01/04  16:18:43  geoff
 * Fix some messages to conform more closely to the System V.2 cron.
 * 
 * Revision 2.5  85/12/21  15:16:57  geoff
 * Add #ifdef's for ABSDIRS, and a bit of code to make relative directories
 * work right.
 * 
 * Revision 2.4  85/11/25  16:54:10  geoff
 * Move the lockout call to after the fork, so we write the correct PID!
 * Call readall once at the end of making the filelist, rather than
 * calling readfile while we build it.  Create the lock file with octal
 * 444 permissions, not decimal 444.
 * 
 * Revision 2.3  85/11/24  16:15:35  geoff
 * Fix some erroneous comments, and delete mention of the 'f' flag in
 * crontab lines.  Add the filelist, usage, and lockout routines.
 * Remove support for filenames on the command line, and call filelist
 * instead.  Use lockout to make sure only one cron is running.
 * Fix a bunch of lint complaints.
 * 
 * Revision 2.2  85/11/05  00:37:58  geoff
 * Define and initialize the new global "prochead".
 * 
 * Revision 2.1  85/11/03  00:11:20  geoff
 * Update the commentary on the file format to reflect the new one.  Add the
 * -n (never really run commands) debug option and support for passing the TZ
 * environment variable to executed commands.
 * 
 * Revision 1.6  85/10/12  01:33:44  geoff
 * Break it up into separate modules
 * 
 * Revision 1.5  85/10/11  21:14:28  geoff
 * Make it compile without the DEBUG option
 * 
 * Revision 1.4  85/10/08  20:33:06  geoff
 * Add fseek before all writes to stderr (cronlog).  Add cf_cwd stuff.
 * Add catchup code.  Clean up some minor stuff.
 * 
 * Revision 1.3  85/10/04  01:11:11  geoff
 * Add symbolic uid/gid, % idiocy, ce_input stuff
 * 
 * Revision 1.2  85/10/03  23:03:06  geoff
 * Finish writing the basic stuff and making it all work
 * 
 * Revision 1.1  85/10/02  15:29:10  geoff
 * Initial revision
 * 
 */

/*
 * Extended version of cron(1), with "catchup" facilities for times when the
 * machine was down.
 *
 * Usage:
 *
 *	cron [-x n] [-f] [-n] [-s interval] [-l logfile] [-t timeout]
 *
 * Argument parsing is done with getopt (3).
 *
 * The -x option specifies a debugging level.  It is only effective if
 * the program was compiled with the DEBUG option.
 *
 * The -f option forces cron to fork into the background, even if -x has
 * been specified.  -f must follow -x.  This is different than running
 * cron in the background with the '&' shell character, because of the
 * way signals are handled.
 *
 * The -n switch suppress the actual running of commands.
 *
 * The default logfile is /usr/lib/cron/log.
 *
 * The -s switch specifies that a sync (2) call is to be done every <interval>
 * seconds.
 *
 * The -t switch specifies a timeout that is to be applied to commands run
 * from user files during 'w' and 's' startup sequences.
 *
 * Cron tables are found in /usr/spool/cron/crontabs.  At jobs are found
 * in /usr/spool/cron/atjobs.  For security, the set-UID and set-GID bits
 * must be on in all cron tables and at jobs, and the execute, sticky,
 * and world write bits must all be turned off.
 *
 * Cron should *not* be run as a set-uid program.  It should be run from
 * /etc/rc as a superuser program.  It will exit immediately if it is not
 * run as superuser (except in debug mode) or if run as a set-uid program.
 *
 * The format of the crontab line is as follows:
 *
 * <flags> <time> <command>
 *
 * <flags> is an extension to the old cron(1M) command line, and is discussed
 * below.
 *
 * <time> is a time, expressed in the following form:
 *
 *	weekday month day year hour:minute
 *
 * where each field can be omitted entirely to indicate all values (except
 * hour:minute;  see below) or can be made up of any number of
 * comma-separated subfields.  Each subfield can be either:
 *
 *	-	A single value, e.g. 20.
 *	or
 *	-	A dash-separated inclusive range of values, e.g., 10-20.
 *		The range can followed by a slash and an increment to
 *		specify only a subset of values within the range.  For
 *		example, 0-59/15 is the same as 0,15,30,45.  Ranges may
 *		wrap around zero if necessary;  for example the range
 *		Sep-Mar/2 selects the odd-numbered fall and winter months.
 *
 * Any complete field specification can be preceded by an exclamation point
 * to indicate all times EXCEPT those selected.  For example, !jun,jul,aug
 * means the same as sep-may.
 *
 * The weekday and month are given symbolically, using the names given in
 * the DAYNAMES and MONTHNAMES configuration variables (in English, the
 * standard three-character abbreviations).  The year must be exactly four
 * digits.
 *
 * The hour and minute must include a colon, even if half is omitted.
 * An omitted hour or minute specification may also be replaced by an
 * asterisk to improve the looks of things.
 *
 * The hour/minute field must appear, even if all values are desired for
 * both.  (Note that this means run every minute, which is probably not
 * very desirable.)  Thus, the minimum time specification is a single colon.
 *
 * As an extra-specially ugly kludge, the hour (NOT the minute) can be
 * followed by "A", "P", "AM", or "PM" to indicate 12-hour time (case is
 * ignored).
 *
 * The command field begins at the first non-blank character after the
 * hour, and continues until the end of the line.  Long commands may be
 * extended over multiple lines by backslashing the newline.  Leading
 * whitespace is stripped from continuation lines;  this may be suppressed
 * by backslashing the first whitespace character of the line.  Pairs of
 * backslashes are always converted to single backslashes, and the sequence
 * '\n' is converted to a newline.  Other backslash combinations (including
 * backslashed whitespace, as long as it isn't at the beginning of a line)
 * are passed through unchanged.
 *
 * The command is passed to the "system" routine for execution.
 *
 * The flags are a sequence of alphabetic or special (but not numeric)
 * characters that specify special processing, and may indicate that
 * additional fields appear after the weekday field:
 *
 *	w	Run this command during the catchup phase of startup (see
 *		below).  Wait for it to finish before continuing with startup.
 *		Note that this can delay entry into the main loop by a
 *		considerable time.  A time limit may be set on commands from
 *		user cronfiles that use this option with the '-t' switch.
 *	&	Run this command in background during the catchup phase, but
 *		do not wait for it to terminate.
 *	s	Run this command as a "sequenced background" command during
 *		the catchup phase.  During catchup, all 's' mode commands
 *		are collected and run in sequence by a background process.
 *		This combines the best features of 'w' and '&':  Sequencing
 *		built into the order of cronfile entries is preserved, and
 *		system load is minimized, but the main loop of cron is
 *		entered quickly.  A time limit may be set on user cronfiles
 *		with the '-t' switch on the command line.
 *
 * The 'w', '&', and 's' flags are mutually exclusive.  The following flags
 * may appear in any combination:
 *
 *	a	During the catchup phase, run this command as many times as
 *		necessary to compensate for the time the system was down.
 *		This command should not be combined with the '&' option, as it
 *		could flood the system with processes.  See below for more
 *		information.  If none of 'w', '&', and 's' is specified with
 *		'a', 'w' is assumed.
 *	u	Run this command under the user ID specified in the first
 *		field following the weekday.  This is equivalent to, but
 *		quicker than, putting 'exec /bin/su userid -c' in front of
 *		the command.  This is only allowed in /usr/lib/crontab.
 *	g	Run this command under the group ID specified in the first
 *		field following the weekday.  This is equivalent to, but
 *		quicker than, putting 'exec /bin/newgrp groupid -c' in front of
 *		the command.  This is only allowed in /usr/lib/crontab.
 *
 * Catchup Mode:
 *
 * The major feature of this version of cron is the presence of 'catchup'
 * mode, which is a feature intended to make UNIX more palatable on small
 * workstations.  Xcron keeps track (to the minute) of when it last ran
 * a command.  When it is restarted after a system shutdown or crash, it
 * enters a 'catchup' mode where it tries to run commands that should have
 * run during the shutdown period.  Thus, for example, if you have a
 * crontab entry that runs at 4 AM on the 30th of the month, and you happen
 * to be away on vacation with your workstation turned off, your command
 * will be run when you get back instead.  You can control all of this with
 * the flags;  a crontab line with no flags will never be executed in
 * catchup mode.
 *
 * During catchup mode, xcron scans the crontab for command lines that
 * should have been run while the system was down.  (The scanning is done
 * in a fashion that will preserve the time-ordering of the commands,
 * regardless of their ordering in the cronfile).  When it finds one, the
 * command is run in one of several fashions, depending on the flags.  The
 * 'w', '&', and 's' flags select whether xcron waits for the command to
 * complete before entering its main loop ('w)', and whether the commands
 * are run all at once in background ('&') or in the order they would have
 * run in had the machine been up ('w' or 's').
 *
 * Normally, a command is run only once during catchup mode, regardless of
 * how many times it would have run had the machine been up.  This is to
 * prevent a useless flood of catchups of programs like the accounting,
 * which will find a null work queue after the first catchup runs.  However,
 * in some cases it may be necessary to actually run a command the correct
 * number of times (e.g., to keep daily accounting reports straight).  The
 * 'a' flag can be used to specify that the command should be run the
 * correct number of times.  This flag is not allowed with the '&' flag in
 * user cronfiles.
 *
 * Sequencing of 'w' and '&' commands in catchup mode is the same as if the
 * machine had been up.  's' commands, however, are separated and run
 * as a group after all 'w' and '&' commands have been processed.
 *
 * Example crontab for "root":
 *
 * #
 * #	Sub-Hourly
 * #
 * -   *:00-59/10 exec nice -20 env LOGNAME=daemon /usr/lib/uucp/uuxqt
 * ug  *:05-04/15 uucp uucp exec /usr/lib/uucp/ckuucp 750 250 800 400 '11 / 10'
 *
 * #
 * #	Hourly
 * #
 *                 *:00 date>/dev/console ; echo " " > /dev/console 
 *                 *:00 /bin/su - adm -c "/usr/lib/acct/ckpacct"
 * s   Mon-Fri  8-17:00 /usr/lib/sa/sa1 1200 3 &
 * s   Sat,Sun     *:00 /usr/lib/sa/sa1 &
 * s   Mon-Fri 18-07:00 /usr/lib/sa/sa1 &
 * ug              *:01 usenet usenet exec /usr1/usenet/lib/sendbnews \
 * 				     trwrb unbnews /usr1/usenet/batch/trwrb.*
 * &ug             *:08 uucp uucp exec /usr/lib/uucp/uudemon.hr > /dev/null
 *
 * #
 * #	Daily
 * #
 * wug               0:00 usenet usenet exec /usr1/usenet/lib/trimlib
 * sug   Sun-Sat/2   0:10 usenet usenet exec /usr1/usenet/lib/expire -e 13
 * s                 1:00 /usr/bin/calendar -
 * s     Mon-Sat     2:00 /usr/lib/acct/dodisk
 * s     Mon-Sat     4:00 /bin/su - adm -c "/usr/lib/acct/runacct 2> \
 * 			    "/usr/adm/acct/nite/fd2log"
 * wug               4:00 uucp uucp exec /usr/lib/uucp/uudemon.day > /dev/null
 * s                 4:10 exec find /usr/preserve -type f -mtime +7 \
 * 			    -a -exec rm -f {} \;
 * s                 4:15 exec find / -name core -print > /tmp/corelist
 * s                 5:15 exec find / -name '#*' -atime +3 -print -exec rm -f {} \;
 * w    Mon-Fri      9:00 /usr/games/turnoff > /dev/null
 * s                10:00 mail geoff < /usr1/usenet/lib/errlog; \
 * 			    cp /dev/null /usr1/usenet/lib/errlog
 * w    Mon-Fri     18:00 /usr/games/turnon > /dev/null
 * s    Mon-Fri     18:05 /bin/su adm -c "/usr/lib/sa/sa2 -s 8:00 \
 * 			    -e 18:01 -i 3600 -ubcwydaqvm &"
 *
 * #
 * #	Weekly
 * #
 * s     Mon         5:00 /etc/errstop 2>/dev/console; \
 * 			    /bin/mv /usr/adm/errfile /usr/adm/oerrfile; \
 * 			    /usr/lib/errdemon 2> /dev/console; \
 * 			    /usr/bin/errpt /usr/adm/oerrfile | /bin/mail adm
 * s     Sun         5:20 /bin/cat /usr/adm/cronlog /usr/adm/OLDcronlog \
 * 			      > /usr/adm/cronlog.wk; \
 * 			    /bin/cp /dev/null /usr/adm/cronlog; \
 * 			    /bin/cp /dev/null /usr/adm/OLDcronlog
 * s     Sun         5:25 /bin/cat /usr/adm/sulog /usr/adm/OLDsulog \
 * 			      > /usr/adm/sulog.wk; \
 * 			    /bin/cp /dev/null /usr/adm/sulog; \
 * 			    /bin/cp /dev/null /usr/adm/OLDsulog
 * wug   Sun         5:30 uucp uucp exec /usr/lib/uucp/uudemon.wk > /dev/null
 * -     Sat Oct-Jun 1982-1990 16:50,55 /bin/echo '=~[2J~[12;20H~[7m' \
 * 				CSO concert '~[m' \
 * 			      | tr '=~' '\\007\\033' > /dev/console
 * 
 * #
 * #	Monthly
 * #
 * s    Jan-Dec   1 5:15 /bin/su - adm -c "/usr/lib/acct/monacct"
 * s    Jan-Dec   2 4:55 exec find /usr/adm -type f -mtime +45 -exec rm -f {} \;
 *
 */

#define GEN_ID				/* Generate source-control ID's */

#include "cron.h"
#include "ndir.h"
#include "patchlevel.h"
/*
 * The following local include files aren't actually used here, but are
 * brought in anyway so their RCS ID's will get compiled in.
 */
#include "pstring.h"
#include <ctype.h>
#include <errno.h>
#include <setjmp.h>
#include <stdio.h>
#include <sys/ioctl.h>
#ifdef HASGETITIMER
#include <sys/time.h>
#endif /* HASGETITIMER */

int		main ();		/* Extended cron(1m) program */
static SIG_TYPE	catchalarm ();		/* Catch alarm signals */
static void	filelist ();		/* Initialize crontab file list */
static void	usage ();		/* Put out a usage message */
static int	lockout ();		/* Lock out other instances of cron */
time_t		lastruntime ();		/* Get time of last run from file */

/*
 * The following structures are used to head up the crontab and cronfile
 * chains;  no other part of them is used.
 */
struct cronent	enthead;		/* Head of cronent chain */
struct cronfile filehead;		/* Head of cronfile chain */
struct childproc prochead;		/* Head of child process list */
struct cronuser	userhead;		/* Head of cronuser chain */

static char *	cronlog = CRONLOG;	/* Log file */
#ifdef DEBUG
short		debug = 0;		/* Nonzero to get debugging output */
static short	dofork = 0;		/* Fork into background even if dbg */
int		reallyrun = 1;		/* Zero to suppress actual cmd execs */
#endif
time_t		start_timeout = 0;	/* Timeout to apply during startup */
time_t		syncinterval = 0;	/* Interval between syncs if NZ */
char *		tzenvvar = NULL;	/* Value of timezone environ var */

extern struct cronfile * addfile ();	/* Add a file to the cronfile list */
#ifndef HASGETITIMER
extern unsigned	alarm ();		/* Set alarm timeout */
#endif /* HASGETITIMER */
extern int	alarm_caught;		/* NZ if an alarm has been caught */
extern jmp_buf	alarm_jb;		/* Jump buffer for catcing alarms */
extern int	alarm_jb_valid;		/* NZ if alarm_jb has valid value */
extern time_t	catchup ();		/* Do any catching-up required */
#ifdef LOC_CRON
extern void	checkloc();
#endif /* LOC_CRON */
extern void	cron ();		/* Execute timed commands */
extern int	errno;			/* System error number */
extern char *	errmsg ();		/* Return an error message string */
extern SIG_TYPE	deathcatch ();		/* Catch serious-death signals */
extern SIG_TYPE	fifocatch ();		/* Catch READFIFO signals */
extern char *	getenv ();		/* Get value of environment var */
#ifdef LOC_CRON
extern SIG_TYPE loccatch ();
#endif /* LOC_CRON */
extern char *	optarg;			/* Pointer to option argument */
extern int	optind;			/* Index of next argv to process */
extern void	readall ();		/* Read all files */
extern void	readlimits ();		/* Read in per-user queue limits */
#ifdef HASGETITIMER
extern int	setitimer ();		/* Set timer value */
#endif /* HASGETITIMER */
extern unsigned	sleep ();		/* Suspend process for a while */ 
extern long	time ();		/* Return time of day */
extern char *	timenonl ();		/* Return ASCII time w/o newline */

int main (argc, argv, envp)			/* System clock daemon */
    reg3 int			argc;		/* Argument count */
    reg1 char *			argv[];		/* Argument vector */
    reg4 char *			envp[];		/* Environment vector */
    {
    reg5 int			kidpid;		/* Process ID of child */
    reg2 int			nextopt;	/* Next option from getopt */
    reg6 FILE *			logfp;		/* Pointer for opening log */
    reg7 time_t			reftime;	/* Time to catch up from */
    char *			revp;		/* Pointer into revision str */
    static char			revstr[] =	/* Revision str for logging */
				  "$Revision: 2799 $";
    int				saveerr;	/* To save errno b4 clobber */
#ifdef HASGETITIMER
    struct itimerval		zerotimer;	/* Timer for removing alarms */
#endif /* HASGETITIMER */


#ifdef DEBUG
    while ((nextopt = getopt (argc, argv, "fnx:s:t:l:")) != EOF)
#else
    while ((nextopt = getopt (argc, argv, "s:t:l:")) != EOF)
#endif
	{
	switch (nextopt)
	    {
#ifdef DEBUG
	    case 'x':				/* Select debug output level */
		debug = atoi (optarg);
		cronlog = NULL;
		break;
	    case 'n':				/* Never actually run stuff */
		reallyrun = 0;
		break;
	    case 'f':				/* Fork to b.g., even though */
		dofork = 1;			/* ..we are debugging */
		break;
#endif
	    case 'l':				/* Set log file */
		cronlog = optarg;
		break;
	    case 's':				/* Sync disks every n sec */
		syncinterval = atoi (optarg);
		break;
	    case 't':				/* Put timeout on startups */
		start_timeout = atoi (optarg);
		break;
	    default:
		usage ();
		return 1;
	    }
	}
    if (optind < argc)
	{
	usage ();
	return 1;
	}
    /*
     * Verify that we were run by root (and not set-UID), and then fork so that
     * we run in the background and are a child of /etc/init.  Also catch
     * all sorts of signals so we can clean up.  All this is only done
     * if debugging is turned off.
     */
#ifdef HASGETITIMER
    zerotimer.it_interval.tv_sec = zerotimer.it_interval.tv_usec = 0;
    zerotimer.it_value.tv_sec = zerotimer.it_value.tv_usec = 0;
    (void) setitimer (ITIMER_REAL, &zerotimer, (struct itimerval*) NULL);
    (void) setitimer (ITIMER_VIRTUAL, &zerotimer, (struct itimerval*) NULL);
    (void) setitimer (ITIMER_PROF, &zerotimer, (struct itimerval*) NULL);
#endif /* HASGETITIMER */
    (void) alarm (0);			/* Make sure we don't have alarms */
#ifdef DEBUG
    if (dofork  ||  debug == 0)
	{
#endif
	if (getuid () != ROOTUID  ||  getgid () != ROOTGID
	  ||  geteuid () != ROOTUID  ||  getegid () != ROOTGID)
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr, "cron must be run by root\n");
	    return 1;
	    }
	while ((kidpid = fork ()) == -1  &&  errno == EAGAIN)
	    (void) sleep (5);
	if (kidpid == -1)
	    (void) fprintf (stderr,
	      "cron:  couldn't fork to background:  %s\n",
	      errmsg (0));
	else if (kidpid != 0)		/* Parent just exits */
	    {
#ifdef DEBUG
	    if (debug > 0)
		{
		(void) fseek (stderr, 0L, 2); /* Make sure we're at EOF */
		(void) fprintf (stderr, "child process ID is %d\n", kidpid);
		}
#endif
	    return 0;
	    }
	(void) signal (SIGHUP, deathcatch); /* Handle some signals cleanly */
	(void) signal (SIGINT, deathcatch); /* ..even if they were ignored! */
	(void) signal (SIGQUIT, deathcatch);
#ifdef HASSETPGRP
#ifdef SETPGRPARGS
	(void) setpgrp (0, getpid ());	/* Dissociate ourselves from TTY */
#else
	(void) setpgrp ();	
#endif
#endif
#ifdef TIOCNOTTY
	/*
	 * This is the BSD way to disassociate from a TTY.
	 */
	if((nextopt = open("/dev/tty", 2)) >= 0)
	    {
	    ioctl (nextopt, TIOCNOTTY, 0);
	    close (nextopt);
	    }
#endif
#ifdef DEBUG
	}
#endif
    /*
     * Make sure our default nice level is zero.
     */
#ifdef RELNICE
    (void) nice (-(MAXNICE - MINNICE + 1));
#endif
    (void) nice ((MAXNICE - MINNICE + 1) / 2);
    /*
     * Change directories to LIBDIR, which is where we normally live.
     * We do this by changing to CRONTABDIR and then to LIBDIR, in case
     * LIBDIR is relative to CRONTABDIR.
     */
    if (chdir (CRONTABDIR) < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  trouble cd'ing to %s: %s\n",
	  CRONTABDIR, errmsg (saveerr));
	return 1;
	}
    if (chdir (LIBDIR) < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  trouble cd'ing to %s: %s\n",
	  LIBDIR, errmsg (saveerr));
	return 1;
	}
    /*
     * Make sure we are the only cron running
     */
    if (lockout ())			/* Make sure we're the only cron */
	{
	(void) fprintf (stderr,
	  "Some other cron must already be running (I couldn't make %s/%s)\n",
	  LIBDIR, LOCKFILE);
	return 1;
	}
    /*
     * Catch the signals we will need to catch
     */
    (void) signal (READFIFO, fifocatch);
    (void) signal (SIGALRM, catchalarm);
    if (signal (SIGINT, SIG_IGN) != SIG_IGN)
	(void) signal (SIGINT, deathcatch);
    if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
	(void) signal (SIGHUP, deathcatch);
    if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
	(void) signal (SIGTERM, deathcatch);
    /*
     * Set a very restrictive umask, so that nobody can spy on us.
     */
    (void) umask (CRONUMASK);
    /*
     * Set up standard input, output, and error for cron programs.
     * Stdin is /dev/null;  stdout and stderr are the cron log file
     * (or /dev/tty if debug mode was requested and no log file was explicitly
     * specified).
     */
#ifdef __linux__
    (void) close (fileno (stdin));
    if (open (NULLINPUT, 0) != fileno (stdin))
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "%s: couldn't open %s for input\n",
	  argv[0], NULLINPUT);
	}
#else /* __linux__ */
    (void) fclose (stdin);
    if (fopen (NULLINPUT, "r") != stdin  ||  fileno (stdin) != 0)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "%s: couldn't open %s for input\n",
	  argv[0], NULLINPUT);
	}
#endif /* __linux__ */
    if (cronlog == NULL  ||  (logfp = fopen (cronlog, "a")) == NULL)
	{
	if (cronlog != NULL)
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "%s:  couldn't open log file %s, logging to existing stderr\n",
	      argv[0]);
	    }
	}
    else
	{
	(void) close (fileno (stderr));
	(void) close (fileno (stdout));
	(void) dup (fileno (logfp));	/* Make new stdout */
	(void) dup (fileno (logfp));	/* Make new stderr */
	(void) fclose (logfp);
	setbuf (stdout, (char *) NULL);
	setbuf (stderr, (char *) NULL);
	}
    /*
     * Set up miscellaneous globals
     */
    filehead.cf_next = filehead.cf_prev = &filehead;
    enthead.ce_next = enthead.ce_prev = &enthead;
    prochead.cp_next = prochead.cp_prev = &prochead;
    userhead.cu_next = userhead.cu_prev = &userhead;
    for (  ;  *envp != NULL;  envp++)	/* Get time zone from environment */
	{
	if (strncmp (*envp, ENVTZ, EL_TZ) == 0)
	    {
	    tzenvvar = *envp;
	    break;
	    }
	}
    revp = strchr (revstr, ':') + 2;
#if 0
    *strchr (revp, ' ') = '\0';
#else /* 0 */
    {
	char *cp = strchr (revp, ' '); /* hack for redhat 6.0 */
	 *cp = '\0';
    }
#endif /* 0 */
    reftime = time ((long *) NULL);
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr,
      "! *** cron version %s patch level %d started ***   pid = %d %s\n",
      revp,
      PATCHLEVEL,
      getpid (),
      timenonl (reftime));
    readlimits ();			/* Read in the per-user limits */
    reftime = lastruntime ();		/* Get time we ran last */
    filelist (reftime);			/* Initialize the file list */
					/* ..this also reads the files */
    reftime = catchup (reftime);	/* Do any catching-up required */
    start_timeout = 0;			/* Disable startup timeout */
    cron (reftime);			/* Execute timed commands */
    /* NOTREACHED */
    }

static SIG_TYPE catchalarm ()		/* Catch alarm signals */
    {
    (void) signal (SIGALRM, catchalarm); /* Re-catch the signal */
    alarm_caught = 1;					    
    if (alarm_jb_valid)
	{
	alarm_jb_valid = 0;
	longjmp (alarm_jb, 1);
	}
    }

static void filelist (reftime)		/* Initialize crontab file list */
    reg3 time_t			reftime; /* Reference time to execute from */
    {
    reg2 struct direct *	dir;	/* Entry just read from directory */
    reg1 DIR *			list;	/* Stream with directory */
    int				saveerr;	/* To save errno b4 clobber */

    /*
     * Now read in the file list by reading CRONTABDIR.
     */
    list = opendir (CRONTABDIR);
    if (list == NULL)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  trouble opening %s: %s\n",
	  CRONTABDIR, errmsg (saveerr));
	return;
	}
    while ((dir = readdir (list)) != NULL)
	{
	if (strcmp (dir->d_name, ".") != 0
	  &&  strcmp (dir->d_name, "..") != 0)
	    (void) addfile (dir->d_name, 0);
	}
    closedir (list);
    
    /*
     * Now read in the at job list in a similar manner.
     */
    list = opendir (ATJOBDIR);
    if (list == NULL)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "cron:  trouble opening %s: %s\n",
	  ATJOBDIR, errmsg (saveerr));
	return;
	}
    while ((dir = readdir (list)) != NULL)
	{
	if (strcmp (dir->d_name, ".") != 0
	  &&  strcmp (dir->d_name, "..") != 0)
	    (void) addfile (dir->d_name, FF_ATJOB);
	}
    closedir (list);
    readall (reftime);
    }

static void usage ()		/* Put out a usage message */
    {
#ifdef DEBUG
    (void) fprintf (stderr,
      "Usage:  cron [-x n] [-f] [-n] [-l logfile] [-s synctime] [-t timeout]\n");
#else
    (void) fprintf (stderr,
      "Usage:  cron [-l logfile] [-s synctime] [-t timeout]\n");
#endif
    }

static int lockout ()		/* Lock out other instances of cron */
    {
    char		lock[sizeof LOCKLINK + 10]; /* Lock file name */
    reg1 FILE *		lockfile; /* Lock file access */
    time_t		now;	/* The current system time */
    int			pid;	/* Pid to be written into the file */
    int			saveerr; /* Place to save errno before clobber */
    struct stat		statbuf; /* Stat buffer for learning about time log */

    pid = getpid ();		/* Get our pid for file names and contents */
    (void) sprintf (lock, LOCKLINK, pid); /* Make file name */
    lockfile = fopen (lock, "w"); /* Make temp lock file */
    if (lockfile == NULL)	/* Failures are same as if cron existed */
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "lockout:  couldn't creat %s: %s\n",
	  lock, errmsg (errno));
	return 1;
	}
    /* Create lock file contents */
    if (chmod (lock, 0444) < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "lockout:  couldn't chmod %s/%s to 0444: %s, continuing\n",
	  LIBDIR, lock, errmsg (saveerr));
	}
    (void) fprintf (lockfile, "%d\n", pid);
    (void) fclose (lockfile);
    /*
     * We have now created the temp file.  Try to link it to the real
     * lock file.  If we succeed, return.  Otherwise try to find out if
     * the process owning the lock still exists, and break the lock if
     * it doesn't.
     */
    while (1)
	{
	if (link (lock, LOCKFILE) >= 0) /* Create real lock file */
	    {
	    (void) unlink (lock);	/* Get rid of temp file */
	    return 0;			/* We're done, and the only cron */
	    }
	lockfile = fopen (LOCKFILE, "r"); /* Try to read the lock file */
	if (lockfile == NULL)
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "lockout:  couldn't open %s/%s for read: %s\n",
	      LIBDIR, LOCKFILE, errmsg (saveerr));
	    (void) unlink (lock);	/* Get rid of temp file */
	    return 1;
	    }
	if (fscanf (lockfile, "%d", &pid) != 1)
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "lockout:  couldn't read PID from %s/%s: %s\n", LIBDIR, LOCKFILE,
	      errmsg (saveerr));
	    (void) fclose (lockfile);
	    (void) unlink (lock);	/* Get rid of temp file */
	    return 1;
	    }
	(void) fclose (lockfile);
	if (kill (pid, 0) < 0  &&  errno == ESRCH) /* Test for dead process */
	    (void) unlink (LOCKFILE);	/* Dead proc, zap it's lock file */
					/* ..and let loop try again */
	else
	    {
	    (void) unlink (lock);	/* Zap our old lock file */
	    (void) time (&now);		/* Guess whether cron's alive */
	    if (stat (TIMELOG, &statbuf) < 0
	      ||  statbuf.st_mtime < now - 2 * MINUPDLAST)
		(void) fprintf (stderr,
		  "lockout:  %s/%s exists, but process %d probably isn't cron\n",
		  LIBDIR, LOCKFILE, pid);
	    return 1;			/* And fail */
	    }
	}
    }

time_t lastruntime ()		/* Get time of last run from record file */
    {
    reg1 int		fd;		/* Fd for creating TIMELOG */
    int			saveerr;	/* To save errno b4 clobber */
    struct stat		statbuf;	/* For getting last-time record */

    if (stat (TIMELOG, &statbuf) < 0)
	{
	if (errno == ENOENT)		/* Nonexistent file? */
	    {
	    fd = creat (TIMELOG, 0644);	/* Create the time log */
	    if (fd < 0)
		{
		saveerr = errno;
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr, "Couldn't create %s: %s\n",
		  TIMELOG, errmsg (saveerr));
		}
	    else
		(void) close (fd);
	    }
	else
	    {
	    saveerr = errno;
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "Couldn't stat %s:  %s\n", TIMELOG, errmsg (saveerr));
	    }
	statbuf.st_mtime = time((long *) NULL);
	}
    return statbuf.st_mtime;
    }

